// save.C    : functions for saving & loading images in different modes
//  a part of PMS-grabber package
// the "GNU Public Lincense" policy applies to all programs of this package
// (c) Wolfgang Koehler, wolf@first.gmd.de, Dec. 1994
//     Kiefernring 15
//     14478 Potsdam, Germany
//     0331-863238

/******************** JPEG COMPRESSION SAMPLE INTERFACE *******************/

struct ppm { // struct for 3byte ppm rgb pixels rgb=888 needed for jpeg
  unsigned char r,g,b;
  inline void boost(unsigned short c565) { // blow up rgb565 -> rgb888 
    r = (c565 >> 11) << 3;
    g = (c565 >> 5) << 2;
    b = c565 << 3;
  }
  inline unsigned short c565() { // return standard 565 color
    unsigned short val = b >> 3;
    val |= (g >> 2) << 5; 
    val |= (r >> 3) << 11;
    return val;
  }
};

#include <stdlib.h>
#include <stdio.h>

extern "C" {
#include "jpeg/jpeglib.h"
}

// to be called after initialization of the cinfo struct 
// write one frame compressed to file stream associated with cinfo
// may be called succesively for more frames
void JPEG_compress(struct jpeg_compress_struct *cinfo, int image_width,
		   unsigned short *image_buffer) {

  jpeg_start_compress(cinfo, TRUE);
  struct ppm ppm_buffer[image_width]; // buffer for one scanline

  /* pointer to JSAMPLE row[s] */
  JSAMPROW pptr = (unsigned char*) ppm_buffer;

  while (cinfo->next_scanline < cinfo->image_height) {
    unsigned short *im_line = image_buffer + image_width*cinfo->next_scanline;
    for (int i=0; i < image_width; i++) ppm_buffer[i].boost(im_line[i]);
    jpeg_write_scanlines(cinfo, &pptr, 1);
  }
  jpeg_finish_compress(cinfo);
}

void compress_jpeg(FILE *fsave, int quality, 
		   int ww, int hh, unsigned short* image_buf) {
  struct jpeg_compress_struct cinfo;
  struct jpeg_error_mgr jerr;

  cinfo.err = jpeg_std_error(&jerr);
  jpeg_create_compress(&cinfo);
  jpeg_stdio_dest(&cinfo, fsave);

  cinfo.image_width = ww; 	/* image width and height, in pixels */
  cinfo.image_height = hh;
  cinfo.input_components = 3;		/* # of color components per pixel */
  cinfo.in_color_space = JCS_RGB; 	/* colorspace of input image */
  jpeg_set_defaults(&cinfo);
  jpeg_set_quality(&cinfo, quality, TRUE /* limit to baseline-JPEG values */);
  JPEG_compress(&cinfo,ww,image_buf);

  jpeg_destroy_compress(&cinfo);
}

// to be called after initialization of the cinfo struct 
// read one frame from file stream associated with cinfo and decompress 
// may be called succesively for more frames of a video
void JPEG_decompress(struct jpeg_decompress_struct *cinfo, 
		     unsigned short *image_buffer) {

  int ww = cinfo->output_width; 
  int row_stride = ww * cinfo->output_components;
  
  /* Make a one-row-high sample array that will go away when done with image */
   JSAMPARRAY buffer = (*cinfo->mem->alloc_sarray)
                       ((j_common_ptr) cinfo, JPOOL_IMAGE, row_stride, 1);

  while (cinfo->output_scanline < cinfo->output_height) {
    jpeg_read_scanlines(cinfo, buffer, 1);
    unsigned short *capline= image_buffer + ww*cinfo->output_scanline;
    ppm* ppline = (ppm*) buffer[0];
    for (int i=0; i < ww; i++) capline[i] = ppline[i].c565();
  }
}
